#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.constants as constants


class ChronologicalReadDatabaseCache:
    DATA_KEY = "data"
    DATA_SORT_KEY = "data_sort_key"
    CHRONO_INDEX_KEY = "chrono_index"

    def __init__(self):
        self.timestamped_sorted_data = {}

    def set(self, values, sort_key, identifiers):
        """
        Set whole cache to later be able to efficiently select it
        :param values: cache values to set
        :param sort_key: key in the values dict to use to chronologically order data
        :param identifiers: identifiers of the given cache. Used to store multiple cache sets
        """
        nested_cache = self.timestamped_sorted_data
        for identifier in identifiers:
            if identifier not in nested_cache:
                nested_cache[identifier] = {}
            nested_cache = nested_cache[identifier]
        data = self._get_cache_data(identifiers)
        data[self.DATA_SORT_KEY] = sort_key
        data[self.DATA_KEY] = sorted(values, key=lambda x: x[sort_key])
        data[self.CHRONO_INDEX_KEY] = 0

    def reset_cached_indexes(self, parent=None):
        """
        Set the cache index of each cached element at 0
        :param parent: current cached element
        """
        for cached_data in (parent or self.timestamped_sorted_data).values():
            if isinstance(cached_data, dict):
                if self.CHRONO_INDEX_KEY in cached_data:
                    cached_data[self.CHRONO_INDEX_KEY] = 0
                else:
                    self.reset_cached_indexes(cached_data)

    def get(self, inferior_timestamp, superior_timestamp, identifiers):
        """
        Returns a cache values
        :param inferior_timestamp: timestamp to start selecting from. Use constants.DEFAULT_IGNORED_VALUE to select all
        :param superior_timestamp: timestamp to stop selecting at. Use constants.DEFAULT_IGNORED_VALUE to select all
        :param identifiers: identifiers of the cache to look into. Used to store multiple cache sets
        """
        cache_data = self._get_cache_data(identifiers)
        # if one timestamp is constants.DEFAULT_IGNORED_VALUE, return every available data from/up to this timestamp
        if inferior_timestamp == constants.DEFAULT_IGNORED_VALUE:
            if superior_timestamp == constants.DEFAULT_IGNORED_VALUE:
                return cache_data[self.DATA_KEY]
            return [
                element
                for element in cache_data[self.DATA_KEY]
                if element[cache_data[self.DATA_SORT_KEY]] <= superior_timestamp
            ]
        if superior_timestamp == constants.DEFAULT_IGNORED_VALUE:
            return [
                element
                for element in cache_data[self.DATA_KEY]
                if element[cache_data[self.DATA_SORT_KEY]] >= inferior_timestamp
            ]
        return self._get_from_time_window(
            cache_data, inferior_timestamp, superior_timestamp
        )

    def _get_from_time_window(self, cache_data, inferior_timestamp, superior_timestamp):
        data = cache_data[self.DATA_KEY]
        data_sort_key = cache_data[self.DATA_SORT_KEY]
        # a specific time window is requested: use the local cache index to generate it faster
        min_index = max_index = None
        # identify the select window considering data are time sorted
        start_index = cache_data[self.CHRONO_INDEX_KEY]
        for index in range(start_index, len(data)):
            element = data[index]
            if element[data_sort_key] >= inferior_timestamp and min_index is None:
                min_index = index
            if element[data_sort_key] > superior_timestamp and max_index is None:
                max_index = index
                # consider that since inferior_timestamp got requested, as this is a chronological database cache,
                # is going only following requests will only be in future times. Therefore previous data can be
                # ignored to avoid iterating over them over and over
                cache_data[self.CHRONO_INDEX_KEY] = min_index
                return data[min_index:max_index]
        if min_index is None:
            return []
        return data[min_index:]

    def has(self, identifiers):
        """
        :param identifiers: identifiers of the cache to look for.
        :return: True if the current identifiers are related to a registered cache
        """
        try:
            self._get_cache_data(identifiers)
            return True
        except KeyError:
            return False

    def _get_cache_data(self, identifiers):
        found_data = self.timestamped_sorted_data
        for identifier in identifiers:
            found_data = found_data[identifier]
        return found_data

    def clear(self):
        """
        Resets the cache database
        """
        self.timestamped_sorted_data = {}



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.*
import json
import pytest

import octobot_commons.profiles as profiles
import octobot_commons.constants as constants
import octobot_commons.enums as enums

from tests.profiles import get_profile_path, profile


@pytest.fixture
def profile_data_dict():
    return {
        'profile_details': {
            'name': 'profile_name 42',
            'id': 'default',
            'version': "42.42.1b",
            'bot_id': "1234-1224-0000"
        },
        'crypto_currencies': [
            {
                'trading_pairs': ['BTC/USDT'],
                'name': 'Bitcoin'
            }
        ], 'exchanges': [
            {
                'exchange_credential_id': '123-plop',
                'internal_name': 'cryptocom'
            }
        ], 'trader': {
            'enabled': True
        }, 'trader_simulator': {
            'enabled': False,
            'starting_portfolio': {
                'BTC': 10,
                'USDT': 1000
            },
            'maker_fees': 0.1,
            'taker_fees': 0.1
        }, 'trading': {
            'reference_market': 'BTC',
            'risk': 0.5
        }, 'tentacles': [
            {
                'name': 'plopEvaluator',
                'config': {},
            },
            {
                'name': 'plopEvaluator',
                'config': {
                    'a': True,
                    'other': {
                        'l': [1, 2],
                        'n': None,
                    }
                },
            },
        ], 'options': {
            'values': {
                'plop_key': 'hola senior'
            }
        }
    }

@pytest.fixture
def min_profile_data_dict():
    return {
        'profile_details': {
            'name': 'min_profile',
        },
        'crypto_currencies': [
            {
                'trading_pairs': ['BTC/USDT'],
            }
        ], 'trading': {
            'reference_market': 'BTC',
        }, 'tentacles': [
            {
                'name': 'plopEvaluator',
                'config': {},
            },
            {
                'name': 'plopEvaluator',
                'config': {
                    'a': True,
                    'other': {
                        'l': [1, 2],
                        'n': None,
                    }
                },
            },
        ], 'options': {
            'values': {
                'plop_key': 'hola !!!',
                'jour': 'nuit',
            }
        }
    }


def test_from_profile(profile):
    profile_data = profiles.ProfileData.from_profile(profile.read_config())
    # check one element per attribute to be sure it's all parsed
    assert profile_data.profile_details.name == "default"
    assert profile_data.crypto_currencies[0].trading_pairs == ['BTC/USDT']
    assert profile_data.exchanges == []
    assert profile_data.trader.enabled is False
    assert profile_data.trader_simulator.enabled is True
    assert profile_data.trader_simulator.starting_portfolio == {'BTC': 10, 'USDT': 1000}
    assert profile_data.trading.risk == 0.5
    assert profile_data.tentacles == []


def test_to_profile(profile):
    profile_data = profiles.ProfileData.from_profile(profile.read_config())
    created_profile = profile_data.to_profile("plop_path")
    # force missing values
    for crypto_data in profile.config[constants.CONFIG_CRYPTO_CURRENCIES].values():
        crypto_data[constants.CONFIG_ENABLED_OPTION] = crypto_data.get(constants.CONFIG_ENABLED_OPTION, True)
    # remove not stored values
    profile.config[constants.CONFIG_EXCHANGES] = {}
    profile.avatar = profile.description = ""
    profile.complexity = enums.ProfileComplexity.MEDIUM
    profile.risk = enums.ProfileRisk.MODERATE
    profile.profile_type = enums.ProfileType.LIVE
    profile.origin_url = None
    # if both parsing and transforming return the same profile as original one, the whole chain works
    profile_dict = profile.as_dict()
    assert profile_dict == created_profile.as_dict()


def test_from_dict(profile_data_dict):
    profile_data = profiles.ProfileData.from_dict(profile_data_dict)
    # check one element per attribute to be sure it's all parsed
    assert profile_data.profile_details.name == "profile_name 42"
    assert profile_data.crypto_currencies[0].trading_pairs == ['BTC/USDT']
    assert profile_data.exchanges[0].exchange_credential_id == "123-plop"
    assert profile_data.exchanges[0].internal_name == "cryptocom"
    assert profile_data.trader.enabled is True
    assert profile_data.trader_simulator.enabled is False
    assert profile_data.trader_simulator.starting_portfolio == {'BTC': 10, 'USDT': 1000}
    assert profile_data.trading.risk == 0.5
    assert profile_data.tentacles[0].name == "plopEvaluator"
    assert profile_data.tentacles[1].config["other"]["l"] == [1, 2]


def test_from_min_dict(min_profile_data_dict):
    profile_data = profiles.ProfileData.from_dict(min_profile_data_dict)
    # check one element per attribute to be sure it's all parsed
    assert profile_data.profile_details.name == "min_profile"
    assert profile_data.crypto_currencies[0].trading_pairs == ['BTC/USDT']
    assert profile_data.crypto_currencies[0].name is None
    assert profile_data.exchanges == []
    assert profile_data.trader.enabled is True
    assert profile_data.trader_simulator.enabled is False
    assert profile_data.trader_simulator.starting_portfolio == {}
    assert profile_data.trading.risk == 1
    assert profile_data.tentacles[0].name == "plopEvaluator"
    assert profile_data.tentacles[1].config["other"]["l"] == [1, 2]
    assert profile_data.options.values["jour"] == "nuit"
    assert profile_data.options.values["plop_key"] == "hola !!!"
    full_profile_data_dict = profile_data.to_dict(include_default_values=True)
    assert len(full_profile_data_dict) > len(min_profile_data_dict)
    profile_data_dict = profile_data.to_dict(include_default_values=False)
    # default values in values but all keys are present
    assert len(profile_data_dict) == len(full_profile_data_dict)


def test_from_dict_objects(profile_data_dict):
    profile_data_objects = profiles.ProfileData.from_dict(profile_data_dict)
    profile_data = profiles.ProfileData.from_dict({
        "profile_details": profile_data_objects.profile_details,
        "crypto_currencies": profile_data_objects.crypto_currencies,
        "exchanges": profile_data_objects.exchanges,
        "trader": profile_data_objects.trader,
        "trader_simulator": profile_data_objects.trader_simulator,
        "trading": profile_data_objects.trading,
        "tentacles": profile_data_objects.tentacles,
        "options": profile_data_objects.options,
    })
    # check one element per attribute to be sure it's all parsed
    assert profile_data.profile_details.version == "42.42.1b"
    assert profile_data.crypto_currencies[0].trading_pairs == ['BTC/USDT']
    assert profile_data.exchanges[0].exchange_credential_id == "123-plop"
    assert profile_data.trader.enabled is True
    assert profile_data.trader_simulator.enabled is False
    assert profile_data.trader_simulator.starting_portfolio == {'BTC': 10, 'USDT': 1000}
    assert profile_data.trading.risk == 0.5
    assert profile_data.tentacles[0].name == "plopEvaluator"
    assert profile_data.tentacles[1].config["other"]["l"] == [1, 2]
    assert profile_data.options.values['plop_key'] == 'hola senior'


def test_to_dict(profile_data_dict):
    profile_data = profiles.ProfileData.from_dict(profile_data_dict)
    # if both parsing and transforming return the same profile as original one, the whole chain works
    assert profile_data_dict == profile_data.to_dict()

    dict_without_default_values = profile_data.to_dict(include_default_values=False)
    assert profile_data_dict != dict_without_default_values

    # ensure no empty elements
    assert len(dict_without_default_values) == len(profile_data_dict)
    for values in dict_without_default_values.values():
        assert len(values)



# pylint: disable=C0116,W0511,R0913
#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import contextlib
import sqlite3
import aiosqlite

import octobot_commons.logging as logging
import octobot_commons.enums as enums
import octobot_commons.errors as errors
import octobot_commons.databases.relational_databases.sqlite.cursor_pool as cursor_pool


class SQLiteDatabase:
    TIMESTAMP_COLUMN = "timestamp"
    DEFAULT_ORDER_BY = TIMESTAMP_COLUMN
    DEFAULT_SORT = enums.DataBaseOrderBy.DESC.value
    DEFAULT_WHERE_OPERATION = "="
    DEFAULT_SIZE = -1
    CACHE_SIZE = 50

    def __init__(self, file_name):
        self.file_name = file_name
        self.logger = logging.get_logger(self.__class__.__name__)

        self.tables = []
        self.cache = {}

        self.connection = None

        # should never be used directly, use async with self.aio_cursor() as cursor: instead
        self._cursor_pool = None

    async def initialize(self):
        try:
            self.connection = await aiosqlite.connect(self.file_name)
            self._cursor_pool = cursor_pool.CursorPool(self.connection)
            await self.__init_tables_list()
        except (sqlite3.OperationalError, sqlite3.DatabaseError) as err:
            raise errors.DatabaseNotFoundError(f"{err} (file: {self.file_name})")

    async def create_index(self, table, columns):
        await self.__execute_index_creation(
            table, "_".join(columns), ", ".join(columns)
        )

    @contextlib.asynccontextmanager
    async def aio_cursor(self) -> sqlite3.Cursor:
        """
        Use this as a context manager to get a free database cursor
        :yield: A free cursor
        :return: None
        """
        async with self._cursor_pool.idle_cursor() as cursor:
            yield cursor.cursor

    async def __execute_index_creation(self, table, name, columns):
        async with self.aio_cursor() as cursor:
            await cursor.execute(
                f"CREATE INDEX index_{table.value}_{name} ON {table.value} ({columns})"
            )

    async def insert(self, table, timestamp, **kwargs):
        if table.value not in self.tables:
            await self.__create_table(table, **kwargs)

        # Insert a row of data
        inserting_values = [f"'{value}'" for value in kwargs.values()]
        await self.__execute_insert(
            table, self.__insert_values(timestamp, ", ".join(inserting_values))
        )

    async def insert_all(self, table, timestamp, **kwargs):
        # TODO refactor with : cursor.executemany("INSERT INTO my_table VALUES (?,?)", values)
        if table.value not in self.tables:
            await self.__create_table(table, **kwargs)

        insert_values = []

        for index, values in enumerate(timestamp):
            # Insert a row of data
            inserting_values = [
                f"'{value if not isinstance(value, list) else value[index]}'"
                for value in kwargs.values()
            ]
            insert_values.append(
                self.__insert_values(values, ", ".join(inserting_values))
            )

        await self.__execute_insert(table, ", ".join(insert_values))

    async def update(self, table, updated_value_by_column, **kwargs):
        # Update a row of data
        updating_values = [
            f"{key} = '{value}'" for key, value in updated_value_by_column.items()
        ]
        await self.__execute_update(
            table,
            ", ".join(updating_values),
            self.__where_clauses_from_kwargs(**kwargs),
        )

    def __insert_values(self, timestamp, inserting_values) -> str:
        return f"({timestamp}, {inserting_values})"

    async def __execute_insert(self, table, insert_items) -> None:
        async with self.aio_cursor() as cursor:
            await cursor.execute(f"INSERT INTO {table.value} VALUES {insert_items}")

        # Save (commit) the changes
        await self.connection.commit()

    async def __execute_update(self, table, update_items, where_clauses) -> None:
        async with self.aio_cursor() as cursor:
            await cursor.execute(
                f"UPDATE {table.value} SET {update_items} WHERE {where_clauses}"
            )

        # Save (commit) the changes
        await self.connection.commit()

    async def select(
        self,
        table,
        size=DEFAULT_SIZE,
        order_by=DEFAULT_ORDER_BY,
        sort=DEFAULT_SORT,
        **kwargs,
    ):
        return await self.__execute_select(
            table=table,
            where_clauses=self.__where_clauses_from_kwargs(**kwargs),
            additional_clauses=self.__select_order_by(order_by, sort),
            size=size,
        )

    async def select_count(self, table, selected_items=None, **kwargs):
        return await self.__execute_select(
            table=table,
            select_items=f"{self.__count(selected_items)}",
            where_clauses=self.__where_clauses_from_kwargs(**kwargs),
        )

    async def select_max(
        self, table, max_columns, selected_items=None, group_by=None, **kwargs
    ):
        return await self.__execute_select(
            table=table,
            select_items=f"{self.__max(max_columns)}"
            f"{', ' if selected_items else ''}"
            f"{self.__selected_columns(selected_items)}",
            where_clauses=self.__where_clauses_from_kwargs(**kwargs),
            group_by=self.__select_group_by(group_by) if group_by else "",
        )

    async def select_min(
        self, table, min_columns, selected_items=None, group_by=None, **kwargs
    ):
        return await self.__execute_select(
            table=table,
            select_items=f"{self.__min(min_columns)}"
            f"{', ' if selected_items else ''}"
            f"{self.__selected_columns(selected_items)}",
            where_clauses=self.__where_clauses_from_kwargs(**kwargs),
            group_by=self.__select_group_by(group_by) if group_by else "",
        )

    async def select_from_timestamp(
        self,
        table,
        timestamps: list,
        operations: list,
        size=DEFAULT_SIZE,
        order_by=DEFAULT_ORDER_BY,
        sort=DEFAULT_SORT,
        **kwargs,
    ):
        timestamps_where_clauses = self.__where_clauses_from_operations(
            keys=[self.TIMESTAMP_COLUMN] * len(timestamps),
            values=timestamps,
            operations=operations,
            should_quote_value=False,
        )
        where_clause = self.__where_clauses_from_kwargs(**kwargs)
        final_where_close = (
            f"{where_clause} AND "
            if where_clause and timestamps_where_clauses
            else where_clause
        )
        final_where_close = f"{final_where_close}{timestamps_where_clauses}"
        return await self.__execute_select(
            table=table,
            where_clauses=final_where_close,
            additional_clauses=self.__select_order_by(order_by, sort),
            size=size,
        )

    async def delete(self, table, **kwargs):
        return await self.__execute_delete(
            table,
            self.__where_clauses_from_kwargs(**kwargs),
        )

    def __where_clauses_from_kwargs(self, should_quote_value=True, **kwargs) -> str:
        return self.__where_clauses_from_operations(
            list(kwargs.keys()),
            list(kwargs.values()),
            [],
            should_quote_value=should_quote_value,
        )

    def __where_clauses_from_operation(
        self, key, value, operation=DEFAULT_WHERE_OPERATION, should_quote_value=True
    ):
        return (
            f"{key} {operation if operation is not None else self.DEFAULT_WHERE_OPERATION} "
            f"{self.__quote_value(value) if should_quote_value else value}"
        )

    def __where_clauses_from_operations(
        self, keys, values, operations, should_quote_value=True
    ):
        return " AND ".join(
            [
                self.__where_clauses_from_operation(
                    keys[i],
                    values[i],
                    operations[i] if len(operations) > i else None,
                    should_quote_value=should_quote_value,
                )
                for i in range(len(keys))
                if values[i] is not None
            ]
        )

    def __select_order_by(self, order_by, sort):
        return (
            f"ORDER BY "
            f"{order_by if order_by is not None else self.DEFAULT_ORDER_BY} "
            f"{sort if sort is not None else self.DEFAULT_SORT}"
        )

    def __select_group_by(self, group_by):
        return f"GROUP BY {group_by}"

    def __quote_value(self, value):
        return f"'{value}'"

    def __max(self, columns):
        return f"MAX({self.__selected_columns(columns)})"

    def __min(self, columns):
        return f"MIN({self.__selected_columns(columns)})"

    def __count(self, columns):
        return f"COUNT({self.__selected_columns(columns)})"

    def __selected_columns(self, columns=None):
        return ",".join(columns) if columns else ""

    async def __execute_select(
        self,
        table,
        select_items="*",
        where_clauses="",
        additional_clauses="",
        group_by="",
        size=DEFAULT_SIZE,
    ):
        try:
            async with self.aio_cursor() as cursor:
                limit_clause = "" if size == self.DEFAULT_SIZE else f"LIMIT {size}"
                await cursor.execute(
                    f"SELECT {select_items} FROM {table.value} "
                    f"{'WHERE' if where_clauses else ''} {where_clauses} "
                    f"{additional_clauses} {limit_clause} {group_by}"
                )
                return await cursor.fetchall()
        except sqlite3.OperationalError as err:
            if not await self.check_table_exists(table):
                raise errors.DatabaseNotFoundError(err)
            self.logger.error(f"An error occurred when executing select : {err}")
        return []

    async def __execute_delete(self, table, where_clauses):
        async with self.aio_cursor() as cursor:
            await cursor.execute(f"DELETE FROM {table.value} WHERE {where_clauses} ")
            # nothing to return, will raise on error

    async def check_table_exists(self, table) -> bool:
        async with self.aio_cursor() as cursor:
            await cursor.execute(
                f"SELECT name FROM sqlite_master WHERE type='table' AND name='{table.value}'"
            )
            return await cursor.fetchall() != []

    async def check_table_not_empty(self, table) -> bool:
        async with self.aio_cursor() as cursor:
            await cursor.execute(f"SELECT count(*) FROM '{table.value}'")
            row_count = await cursor.fetchone()
            return row_count[0] != 0

    async def __create_table(
        self, table, with_index_on_timestamp=True, **kwargs
    ) -> None:
        try:
            columns: list = list(kwargs.keys())
            async with self.aio_cursor() as cursor:
                await cursor.execute(
                    f"CREATE TABLE {table.value} ({self.TIMESTAMP_COLUMN} datetime, "
                    f"{' text, '.join(col for col in columns)})"
                )

            if with_index_on_timestamp:
                await self.create_index(table, [self.TIMESTAMP_COLUMN])

                for i in range(1, round(len(columns) / 2) + 1):
                    await self.create_index(
                        table,
                        [self.TIMESTAMP_COLUMN] + [columns[u] for u in range(0, i)],
                    )

        except sqlite3.OperationalError:
            self.logger.error(f"{table} already exists")
        finally:
            self.tables.append(table.value)

    async def __init_tables_list(self):
        async with self.aio_cursor() as cursor:
            await cursor.execute("SELECT name FROM sqlite_master WHERE type='table'")
            self.tables = [res[0] for res in await cursor.fetchall()]

    async def stop(self):
        try:
            if self._cursor_pool is not None:
                await self._cursor_pool.close()
        finally:
            if self.connection is not None:
                conn = self.connection
                self.connection = None
                await conn.close()


@contextlib.asynccontextmanager
async def new_sqlite_database(file_path):
    local_database = SQLiteDatabase(file_path)
    try:
        await local_database.initialize()
        yield local_database
    finally:
        await local_database.stop()



# pylint: disable=W0703
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.databases.run_databases.file_system_run_databases_pruner as file_system_run_databases_pruner


def run_databases_pruner_factory(run_databases_identifier, max_db_size):
    """
    :return: A RunDatabasesPruner instance
    """
    if run_databases_identifier.database_adaptor.is_file_system_based():
        return file_system_run_databases_pruner.FileSystemRunDatabasesPruner(
            run_databases_identifier,
            max_db_size,
        )
    raise NotImplementedError("Only file system based database pruner is implemented")



# pylint: disable=W0613, W0703, W0719
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import shutil
import os
import json
import octobot_commons.logging as logging
import octobot_commons.constants as commons_constants
import octobot_commons.configuration.fields_utils as fields_utils
import octobot_commons.json_util as json_util


LOGGER_NAME = "ConfigFileManager"


def get_user_config() -> str:
    """
    Return user config path
    :return: user config path
    """
    return os.path.join(commons_constants.USER_FOLDER, commons_constants.CONFIG_FILE)


def load(config_file, should_raise=True, fill_missing_fields=False) -> dict:
    """
    Load a config from a config_file
    :param config_file: the config file path
    :param should_raise: if error should be raised
    :param fill_missing_fields: if missing fields should be filled
    :return: the loaded config
    """
    logger = logging.get_logger(LOGGER_NAME)
    basic_error = "Error when load config file {0}".format(config_file)
    try:
        config = json_util.read_file(config_file)
        return config
    except ValueError as value_error:
        error_str = f"{basic_error} : json decoding failed ({value_error})"
        if should_raise:
            raise Exception(error_str)
        logger.error(error_str)
    except IOError as io_error:
        error_str = f"{basic_error} : file opening failed ({io_error})"
        if should_raise:
            raise Exception(error_str)
        logger.error(error_str)
    except Exception as global_exception:
        error_str = f"{basic_error} : {global_exception}"
        if should_raise:
            raise Exception(error_str)
        logger.error(error_str)
    return None


def dump(
    config_file,
    config,
    temp_restore_config_file=commons_constants.TEMP_RESTORE_CONFIG_FILE,
    schema_file=None,
) -> None:
    """
    Save a json config
    :param config_file: the config file path
    :param config: the json config
    :param temp_restore_config_file: the temporary config file
    :param schema_file: path to the json schema to validate the updated config
    """
    try:
        # prepare a restoration config file
        prepare_restore_file(temp_restore_config_file, config_file)

    # when failing to create the restore config
    except Exception as err:
        error_details = (
            f"Failed to create the backup configuration file. Is your {commons_constants.USER_FOLDER} "
            f"folder accessible ? : {err} ({err.__class__.__name__})"
        )
        logging.get_logger(LOGGER_NAME).error(error_details)
        raise err.__class__(error_details) from err
    try:
        new_content = jsonify_config(config)

        # edit the config file
        with open(config_file, "w") as cg_file:
            cg_file.write(new_content)

        if schema_file is not None:
            # check if the new config file is correct
            check_config(config_file, schema_file)

        # remove temp file
        remove_restore_file(temp_restore_config_file)

    # when failing to restore the previous config
    except Exception as global_exception:
        logging.get_logger(LOGGER_NAME).error(
            f"Save config failed : {global_exception}"
        )
        restore(temp_restore_config_file, config_file)
        raise global_exception


def check_config(config_file, schema_file) -> None:
    """
    Check a config file
    :param config_file: the config file path
    :param schema_file: path to the json schema to validate the updated config
    """
    try:
        json_util.validate(load(config_file=config_file), schema_file=schema_file)
    except Exception as global_exception:
        raise global_exception


def jsonify_config(config) -> str:
    """
    Jsonify a config
    :param config: the config
    :return: the jsonified config
    """
    # check exchange keys encryption
    for exchange, exchange_config in config[commons_constants.CONFIG_EXCHANGES].items():
        try:
            for key in commons_constants.CONFIG_EXCHANGE_ENCRYPTED_VALUES:
                handle_encrypted_value(key, exchange_config)
        except Exception:
            config[commons_constants.CONFIG_EXCHANGES][exchange] = {
                key: "" for key in commons_constants.CONFIG_EXCHANGE_ENCRYPTED_VALUES
            }

    return dump_formatted_json(config)


def handle_encrypted_value(value_key, config_element, verbose=False) -> bool:
    """
    Handle encrypted value
    :param value_key: the value key
    :param config_element: the config element
    :param verbose: if verbosity is enabled
    :return: True if the value can be decrypted
    """
    if value_key in config_element:
        key = config_element[value_key]
        if not fields_utils.has_invalid_default_config_value(key):
            try:
                fields_utils.decrypt(key, silent_on_invalid_token=True)
                return True
            except Exception:
                config_element[value_key] = fields_utils.encrypt(key).decode()
                if verbose:
                    logging.get_logger(LOGGER_NAME).warning(
                        f"Non encrypted secret info found in config ({value_key}): replaced "
                        f"value with encrypted equivalent."
                    )
                return False
    return True


def prepare_restore_file(restore_file, current_config_file) -> None:
    """
    Prepare a config restoration file
    :param restore_file: the restoration file
    :param current_config_file: the file to be restored
    """
    shutil.copy(current_config_file, restore_file)


def remove_restore_file(restore_file) -> None:
    """
    Remove a restore file
    :param restore_file: the restore file path
    """
    os.remove(restore_file)


def restore(restore_file, target_file) -> None:
    """
    Restore a config file from a saved file
    :param restore_file: the restore file path
    :param target_file: the target file path
    """
    shutil.copy(restore_file, target_file)


def dump_formatted_json(json_data) -> str:
    """
    The dumped json data
    :param json_data: the json data to be dumped
    :return: the dumped json data
    """
    return json.dumps(json_data, indent=4, sort_keys=True)
